#ProcessorEnvironment.py
'''Functions to process the environment component.'''

import datetime as dt
import json
import logging
import os
import shutil
from glob import glob
from pathlib import Path

import pandas as pd

# from coordinator.env_suit import env_suit_pipeline
from ews.coordinator.processor_base import ProcessorBase
from ews.coordinator.utils.processor_utils import get_only_existing_globs, get_input_data
from ews.postprocessing.environmental_suitability.env_suit_post_processor import EnvSuitPostProcessor
from ews.postprocessing.utils.disease_info import EnvSuitDiseaseInfo
from ews.met_processing.common.params_file_parser import ParamsFileParser


logger = logging.getLogger(__name__)

class ProcessorEnvironment(ProcessorBase):

    def process_pre_job(self, args) -> bool:
        return True

    def process_in_job(self, job_path, status, configjson, component) -> object:
        return self.process_in_job_env2_0(job_path, status, configjson, component)


    def process_post_job(self, job_path, configjson):
        self.process_ews_plotting(job_path, configjson)


    def __init__(self) -> None:
        super().__init__()

    def process_in_job_env2_0(self, jobPath,status,config,component) -> dict:
        '''Download met data from remote, prepare data, and run :class:`EnvSuitPipeline` pipeline.'''
    
        logger.info('started process_in_job_env2_0()')
    
        # Initialise output variable
        proc_out = {}
        # Output files available for upload    
        proc_out['output'] = None
        # Processing files available for clearing
        proc_out['clearup'] = []

        logger.debug('Copying file from remote server to job directory')
    
        data_result = get_input_data(jobPath, config, component)

        if data_result['output'] is not None:
            proc_out['output'] = data_result['output']

        proc_out['clearup'] += data_result['clearup']
    
        region = config['RegionName']

        # pipeline_config = config["Environment"]
        try:
            logger.info(f"Calling environmental suitability 2.0 for {region}")
            self.run_pipeline(config, region, config["StartString"])
        except:
            logger.exception(f"Some failure when running EnvSuitPipeline.py")
            raise
    
        logger.info('Finished running environmental suitability 2.0')
    
        # TODO: Check that the output appears as expected

        return proc_out


    @staticmethod
    def loadConfig(configFile):
        '''Load a json config file.'''

        logger.info(f"Loading config file: {configFile}")

        try:
            with open(configFile) as config_file:
                config = json.load(config_file)
        except:
            logger.exception(f"Failure opening config file {configFile}")
            raise

        return config


    @staticmethod
    def getParameter(config, parameter):
        '''Get a parameter from a config.'''

        logger.debug(f"Getting {parameter} from the config")

        try:
            result = config[parameter]
        except:
            logger.exception(f"No {parameter} in the config file!")
            raise

        return result


    @staticmethod
    def generate_temporal_points(file, datestr, timeresolution, nDaysForecast):
        '''Create the time series excluding the end time.'''

        datenow = dt.datetime(int(datestr[:4]), int(datestr[4:6]), int(datestr[6:8]), 0, 0)

        starttime = datenow + dt.timedelta(hours = timeresolution)
        endtime = datenow + dt.timedelta(days = nDaysForecast) + dt.timedelta(hours = timeresolution)
        timediff = dt.timedelta(hours = timeresolution)

        timeseries = []
        curr = starttime

        outfile = open(file, "w")
        # Generate the time series excluding the end time.
        while curr < endtime:
            timeseries.append(curr)
            outDateString = "{}\n".format(curr.strftime("%Y%m%d%H%M"))
            outfile.write(outDateString)
            curr += timediff

        outfile.close()
        return outfile


    @staticmethod
    def clean(workPath):
        '''Clean temporary files and folders from the working directory.'''

        try:
            logger.info(f"Clean temporary files and folders from {workPath}")
            shutil.rmtree(workPath + 'extraction', ignore_errors = True)
            shutil.rmtree(workPath + 'chunks', ignore_errors = True)
            shutil.rmtree(workPath + 'post_processing', ignore_errors = True)
        except:
            logger.exception(f"Some failure when cleaning work directory", exc_info = True)
            raise

        return


    @staticmethod
    def run_post_processing(run_params: dict, sys_params: dict, processor_name: str):
        '''Run post processing with the :class:`met_processing.runner.common.job_runner` package.'''

        logger.info(f"Running post-processing.")
        from ews.met_processing.runner import job_runner


        job_runner.run_post_processing(run_params, sys_params, processor_name)
        logger.info('Data post processing is completed')


    @staticmethod
    def run_pipeline(pipeline_config, region, dateString, prevent_overwrite = True):
        '''
        Run the whole :class:`met_processing` pipeline for environmental suitability.
        '''
        # The prevent_overwrite parameter can be set to False if you want to re-run a job in-place.

        # Get parameters from the config
        resourcesPath = ProcessorEnvironment.getParameter(pipeline_config, 'CONFIGS_PATH')
        assetsPath = ProcessorEnvironment.getParameter(pipeline_config, 'ASSETS_PATH')
        workPath = ProcessorEnvironment.getParameter(pipeline_config, 'WORK_PATH') + 'ENVIRONMENT_2.0_' + dateString + '/'
        if not os.path.exists(workPath):
            os.makedirs(workPath)

        inPath = ProcessorEnvironment.getParameter(pipeline_config, 'INPUT_PATH')
        outPath = ProcessorEnvironment.getParameter(pipeline_config, 'OUTPUT_PATH')

        runType = ProcessorEnvironment.getParameter(pipeline_config, 'RUN_TYPE')
        nDayExtraction = ProcessorEnvironment.getParameter(pipeline_config, 'EXTRACTION_DAYS')
        nDayForecast = ProcessorEnvironment.getParameter(pipeline_config, 'FORECAST_DAYS')

        sys_config_file = ProcessorEnvironment.getParameter(pipeline_config, 'SYS_CONFIG')
        sys_config = ProcessorEnvironment.loadConfig(sys_config_file)

        templateName = 'template_' + runType + '_config.json'
        template_configFile = resourcesPath + '/met_extractor/' + templateName
        config = ProcessorEnvironment.loadConfig(template_configFile)

        # Before writing any files, check the output path doesn't exist already
        # We might expect outPath to exist already, but not the processed subfolder
        region_outPath = os.path.join(outPath, 'ENVIRONMENT_2.0_' + dateString, 'processed', region)
        if prevent_overwrite: assert not os.path.exists(region_outPath)

        # Generate extraction (input) and output temporal points files
        timeresolution = 3  # hours

        extraction_temporal_points_file = workPath + 'extraction_temporal_points.csv'
        try:
            logger.info(f"Generate extraction temporal points to: {extraction_temporal_points_file}")
            if prevent_overwrite: assert not os.path.exists(extraction_temporal_points_file)
            ProcessorEnvironment.generate_temporal_points(extraction_temporal_points_file, dateString, timeresolution, nDayExtraction)
        except:
            logger.exception(f"Some failure when generate {extraction_temporal_points_file}", exc_info = True)
        # extraction_temporal_points = pd.read_csv(extraction_temporal_points_file)

        output_temporal_points_file = workPath + 'output_temporal_points.csv'
        try:
            logger.info(f"Generate output temporal points to: {output_temporal_points_file}")
            if prevent_overwrite: assert not os.path.exists(output_temporal_points_file)
            ProcessorEnvironment.generate_temporal_points(output_temporal_points_file, dateString, timeresolution, nDayForecast)
        except:
            logger.exception(f"Some failure when generate {output_temporal_points_file}", exc_info = True)
        output_temporal_points = pd.read_csv(output_temporal_points_file)
        temporal_dim = output_temporal_points.shape[0]

        # Modify run_config
        config[ParamsFileParser.TIMEPOINTS_FILE_KEY] = extraction_temporal_points_file
        config[ParamsFileParser.OUTPUT_DIR_KEY] = workPath
        # config[ParamsFileParser.SPATIAL_POINTS_FILE_KEY] = input_spatial_points_file

        config['FIELD_NAME_CONSTANTS_PATH'] = ProcessorEnvironment.getParameter(pipeline_config, 'FIELD_NAME_CONSTANTS')

        if (runType == 'operational'):
            config['NCDF_DIR_PATH'] = inPath + 'ENVIRONMENT_2.0_' + dateString + '/NAME_Met_as_netcdf/'
        else:
            config['NCDF_DIR_PATH'] = inPath

        ## START RUNS ####################

        N_processor = range(len(config['POST_PROCESSING']['PROCESSORS']))
        logger.info(f"Find {N_processor} processor")

        for p in N_processor:
            processor_name = config['POST_PROCESSING']['PROCESSORS'][p]['PROCESSOR_NAME']
            if processor_name != 'RIE':
                config['POST_PROCESSING']['PROCESSORS'][p]['TIMEPOINTS_FILE_PATH'] = extraction_temporal_points_file

                logger.info(f"Starting {processor_name} post processor ---------------------------------")
                ProcessorEnvironment.run_post_processing(config, sys_config, processor_name)

                # run_merger(config, sys_config, processor_name)
            else:
                strains = ProcessorEnvironment.getParameter(pipeline_config, 'STRAINS')

                config['POST_PROCESSING']['PROCESSORS'][p]['TIMEPOINTS_FILE_PATH'] = output_temporal_points_file

                for strain in strains:
                    # Modify strain specific suitability parts of the config
                    if pipeline_config['PARAMS'][strain]['future_steps'] > 0:
                        for i in range(len(config['POST_PROCESSING']['PROCESSORS'][p]['FUTURE_FIELDS'])):
                            config['POST_PROCESSING']['PROCESSORS'][p]['FUTURE_FIELDS'][i]['DURATION'] = \
                            pipeline_config['PARAMS'][strain]['future_steps']
                    else:
                        for i in range(len(config['POST_PROCESSING']['PROCESSORS'][p]['FUTURE_FIELDS'])):
                            config['POST_PROCESSING']['PROCESSORS'][p]['FUTURE_FIELDS'][i]['ENABLED'] = "FALSE"

                    config['POST_PROCESSING']['PROCESSORS'][p]['PARAMS']['suitability_modules'] = \
                    pipeline_config['PARAMS'][strain]['suitability_modules']
                    config['POST_PROCESSING']['PROCESSORS'][p]['PARAMS']['thresholds'] = \
                    pipeline_config['PARAMS'][strain]['thresholds']

                    logger.info(f"Starting {strain} suitability ---------------------------------")
                    envSuitPath = workPath + 'post_processing/RIE/'
                    ProcessorEnvironment.run_post_processing(config, sys_config, processor_name)

                    # run_merger(config, sys_config, processor_name)

                    resultFile = envSuitPath + 'RIE.nc'
                    strain_outPath = os.path.join(region_outPath, strain)
                    strain_outFile = strain_outPath + '/RIE_value.nc'

                    if not os.path.exists(strain_outPath):
                        os.makedirs(strain_outPath)

                    shutil.copy(resultFile, strain_outFile)

                    # If we have set the flag to write the outputs as csvs, copy them to the processed dir
                    result_csv_file = envSuitPath + 'RIE.csv'
                    if os.path.isfile(result_csv_file):
                        strain_outFile = strain_outPath + '/RIE_value.csv'
                        shutil.copy(result_csv_file, strain_outFile)

                    logger.info(f"{strain} result successfully created and moved to {strain_outPath}/")

        logger.info('SUCCESSFULLY FINISHED')

    @staticmethod
    def process_ews_plotting(jobPath, config) -> [str]:
        '''Configures the plotting arguments and calls EWS-plotting as a python module.
        Returns a list of output files for transfer.'''
    
        logger.info('started process_EWS_plotting_env2_0()')
    
        main_region = config['RegionName']
    
        input_dir = f"{jobPath}/processed/{main_region}"
    
        subregions = config['SubRegionNames']
    
        ews_plotting_output_globs = []
    
        # work on each region
        for region in subregions:
    
            output_dir = f"{jobPath}/plotting/{region.lower()}"
            rie_template_dir = input_dir + "/{DISEASE_DIR}/RIE_value.nc"

            Path(output_dir).mkdir(parents=True, exist_ok=True)
    
            sys_config = config['EWS-Plotting']['SysConfig']
            run_config = config['EWS-Plotting']['RunConfig']
            chart_config = config['EWS-Plotting'][region]['ChartConfig']
            filter_for_country = config['EWS-Plotting'][region]['FilterForCountry']
    
            # Note that this runs all disease types available
    
            logger.info(f"Running EWS-Plotting with the following configs:\n{sys_config}\n{run_config}\n{chart_config}")
    
            env_suit_processor = EnvSuitPostProcessor()
            env_suit_processor.set_param_config_files(sys_params_file_arg = sys_config,
                                                      chart_params_file_arg = chart_config,
                                                      run_params_file_arg = run_config,
                                                      es_output_dir_arg = output_dir,
                                                      issue_date_arg = config['StartString'],
                                                      disease_nc_template_arg = rie_template_dir)
    
            env_suit_processor.run_params.FILTER_FOR_COUNTRY = (filter_for_country.upper() == "TRUE")
            env_suit_processor.run_params.WRITE_CSVS = False

            # Include further diseases in plotting. In this case the irrigated suitabilite for the rusts.
            # TODO: move this part out into a config
            extra_diseases = [
                EnvSuitDiseaseInfo("Stem rust temp-only", "stem_rust_temponly", config['StartString'], "StemRust_TempOnly", rie_template_dir),
                EnvSuitDiseaseInfo("Leaf rust temp-only", "leaf_rust_temponly", config['StartString'], "LeafRust_TempOnly", rie_template_dir),
                EnvSuitDiseaseInfo("Stripe rust temp-only", "stripe_temponly", config['StartString'], "StripeRust_TempOnly", rie_template_dir),
            ]
    
            env_suit_processor.add_diseases(diseases=extra_diseases)
    
            env_suit_processor.process()
    
            # check the output
            ews_plotting_output_dir = f"{output_dir}/images/"

            ews_plotting_output_globs = [f"{ews_plotting_output_dir}*"]
    
        # check the output
        ews_plotting_output_globs = get_only_existing_globs(ews_plotting_output_globs,inplace=False)
    
        # check there is some output from EWS-plotting
        if not ews_plotting_output_globs:
            logger.error('EWS-Plotting did not produce any output')
            raise RuntimeError
    
        # provide list for transfer
        ews_plotting_outputs: [str] = sorted([file for glob_str in ews_plotting_output_globs for file in glob(glob_str)])
    
        return ews_plotting_outputs


if __name__ == '__main__':
    processor = ProcessorEnvironment()
    processor.run_processor("Environment", "ENVIRONMENT_2.0")
